# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: datapath.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0e\x64\x61tapath.proto\x12\x14\x63om.deno.kv.datapath\"?\n\x0cSnapshotRead\x12/\n\x06ranges\x18\x01 \x03(\x0b\x32\x1f.com.deno.kv.datapath.ReadRange\"\xd9\x01\n\x12SnapshotReadOutput\x12\x35\n\x06ranges\x18\x01 \x03(\x0b\x32%.com.deno.kv.datapath.ReadRangeOutput\x12\x15\n\rread_disabled\x18\x02 \x01(\x08\x12#\n\x1bread_is_strongly_consistent\x18\x04 \x01(\x08\x12\x38\n\x06status\x18\x08 \x01(\x0e\x32(.com.deno.kv.datapath.SnapshotReadStatusJ\x04\x08\x03\x10\x04J\x04\x08\x05\x10\x06J\x04\x08\x06\x10\x07J\x04\x08\x07\x10\x08\"G\n\tReadRange\x12\r\n\x05start\x18\x01 \x01(\x0c\x12\x0b\n\x03\x65nd\x18\x02 \x01(\x0c\x12\r\n\x05limit\x18\x03 \x01(\x05\x12\x0f\n\x07reverse\x18\x04 \x01(\x08\"@\n\x0fReadRangeOutput\x12-\n\x06values\x18\x01 \x03(\x0b\x32\x1d.com.deno.kv.datapath.KvEntry\"\x9e\x01\n\x0b\x41tomicWrite\x12+\n\x06\x63hecks\x18\x01 \x03(\x0b\x32\x1b.com.deno.kv.datapath.Check\x12\x31\n\tmutations\x18\x02 \x03(\x0b\x32\x1e.com.deno.kv.datapath.Mutation\x12/\n\x08\x65nqueues\x18\x03 \x03(\x0b\x32\x1d.com.deno.kv.datapath.Enqueue\"\x7f\n\x11\x41tomicWriteOutput\x12\x37\n\x06status\x18\x01 \x01(\x0e\x32\'.com.deno.kv.datapath.AtomicWriteStatus\x12\x14\n\x0cversionstamp\x18\x02 \x01(\x0c\x12\x15\n\rfailed_checks\x18\x04 \x03(\rJ\x04\x08\x03\x10\x04\"*\n\x05\x43heck\x12\x0b\n\x03key\x18\x01 \x01(\x0c\x12\x14\n\x0cversionstamp\x18\x02 \x01(\x0c\"\xcb\x01\n\x08Mutation\x12\x0b\n\x03key\x18\x01 \x01(\x0c\x12,\n\x05value\x18\x02 \x01(\x0b\x32\x1d.com.deno.kv.datapath.KvValue\x12\x39\n\rmutation_type\x18\x03 \x01(\x0e\x32\".com.deno.kv.datapath.MutationType\x12\x14\n\x0c\x65xpire_at_ms\x18\x04 \x01(\x03\x12\x0f\n\x07sum_min\x18\x05 \x01(\x0c\x12\x0f\n\x07sum_max\x18\x06 \x01(\x0c\x12\x11\n\tsum_clamp\x18\x07 \x01(\x08\"N\n\x07KvValue\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\x12\x35\n\x08\x65ncoding\x18\x02 \x01(\x0e\x32#.com.deno.kv.datapath.ValueEncoding\"r\n\x07KvEntry\x12\x0b\n\x03key\x18\x01 \x01(\x0c\x12\r\n\x05value\x18\x02 \x01(\x0c\x12\x35\n\x08\x65ncoding\x18\x03 \x01(\x0e\x32#.com.deno.kv.datapath.ValueEncoding\x12\x14\n\x0cversionstamp\x18\x04 \x01(\x0c\"f\n\x07\x45nqueue\x12\x0f\n\x07payload\x18\x01 \x01(\x0c\x12\x13\n\x0b\x64\x65\x61\x64line_ms\x18\x02 \x01(\x03\x12\x1b\n\x13keys_if_undelivered\x18\x03 \x03(\x0c\x12\x18\n\x10\x62\x61\x63koff_schedule\x18\x04 \x03(\r\"5\n\x05Watch\x12,\n\x04keys\x18\x01 \x03(\x0b\x32\x1e.com.deno.kv.datapath.WatchKey\"{\n\x0bWatchOutput\x12\x38\n\x06status\x18\x01 \x01(\x0e\x32(.com.deno.kv.datapath.SnapshotReadStatus\x12\x32\n\x04keys\x18\x02 \x03(\x0b\x32$.com.deno.kv.datapath.WatchKeyOutput\"\x17\n\x08WatchKey\x12\x0b\n\x03key\x18\x01 \x01(\x0c\"Z\n\x0eWatchKeyOutput\x12\x0f\n\x07\x63hanged\x18\x01 \x01(\x08\x12\x37\n\x10\x65ntry_if_changed\x18\x02 \x01(\x0b\x32\x1d.com.deno.kv.datapath.KvEntry*N\n\x12SnapshotReadStatus\x12\x12\n\x0eSR_UNSPECIFIED\x10\x00\x12\x0e\n\nSR_SUCCESS\x10\x01\x12\x14\n\x10SR_READ_DISABLED\x10\x02*\x80\x01\n\x0cMutationType\x12\x11\n\rM_UNSPECIFIED\x10\x00\x12\t\n\x05M_SET\x10\x01\x12\x0c\n\x08M_DELETE\x10\x02\x12\t\n\x05M_SUM\x10\x03\x12\t\n\x05M_MAX\x10\x04\x12\t\n\x05M_MIN\x10\x05\x12#\n\x1fM_SET_SUFFIX_VERSIONSTAMPED_KEY\x10\t*I\n\rValueEncoding\x12\x12\n\x0eVE_UNSPECIFIED\x10\x00\x12\t\n\x05VE_V8\x10\x01\x12\x0b\n\x07VE_LE64\x10\x02\x12\x0c\n\x08VE_BYTES\x10\x03*v\n\x11\x41tomicWriteStatus\x12\x12\n\x0e\x41W_UNSPECIFIED\x10\x00\x12\x0e\n\nAW_SUCCESS\x10\x01\x12\x14\n\x10\x41W_CHECK_FAILURE\x10\x02\x12\x15\n\x11\x41W_WRITE_DISABLED\x10\x05\"\x04\x08\x03\x10\x03\"\x04\x08\x04\x10\x04\"\x04\x08\x06\x10\x06\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'datapath_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _globals['_SNAPSHOTREADSTATUS']._serialized_start=1601
  _globals['_SNAPSHOTREADSTATUS']._serialized_end=1679
  _globals['_MUTATIONTYPE']._serialized_start=1682
  _globals['_MUTATIONTYPE']._serialized_end=1810
  _globals['_VALUEENCODING']._serialized_start=1812
  _globals['_VALUEENCODING']._serialized_end=1885
  _globals['_ATOMICWRITESTATUS']._serialized_start=1887
  _globals['_ATOMICWRITESTATUS']._serialized_end=2005
  _globals['_SNAPSHOTREAD']._serialized_start=40
  _globals['_SNAPSHOTREAD']._serialized_end=103
  _globals['_SNAPSHOTREADOUTPUT']._serialized_start=106
  _globals['_SNAPSHOTREADOUTPUT']._serialized_end=323
  _globals['_READRANGE']._serialized_start=325
  _globals['_READRANGE']._serialized_end=396
  _globals['_READRANGEOUTPUT']._serialized_start=398
  _globals['_READRANGEOUTPUT']._serialized_end=462
  _globals['_ATOMICWRITE']._serialized_start=465
  _globals['_ATOMICWRITE']._serialized_end=623
  _globals['_ATOMICWRITEOUTPUT']._serialized_start=625
  _globals['_ATOMICWRITEOUTPUT']._serialized_end=752
  _globals['_CHECK']._serialized_start=754
  _globals['_CHECK']._serialized_end=796
  _globals['_MUTATION']._serialized_start=799
  _globals['_MUTATION']._serialized_end=1002
  _globals['_KVVALUE']._serialized_start=1004
  _globals['_KVVALUE']._serialized_end=1082
  _globals['_KVENTRY']._serialized_start=1084
  _globals['_KVENTRY']._serialized_end=1198
  _globals['_ENQUEUE']._serialized_start=1200
  _globals['_ENQUEUE']._serialized_end=1302
  _globals['_WATCH']._serialized_start=1304
  _globals['_WATCH']._serialized_end=1357
  _globals['_WATCHOUTPUT']._serialized_start=1359
  _globals['_WATCHOUTPUT']._serialized_end=1482
  _globals['_WATCHKEY']._serialized_start=1484
  _globals['_WATCHKEY']._serialized_end=1507
  _globals['_WATCHKEYOUTPUT']._serialized_start=1509
  _globals['_WATCHKEYOUTPUT']._serialized_end=1599
# @@protoc_insertion_point(module_scope)
